;------------------ Threading methods for an interpreter in C -----------------

(provide 'threading)

"
------ Parameters ------
This module requires the following parameters (and provides default values)
1) The way instructions are executed as __INSTRUCTION_TYPE__
2) The way instructions are coded into tokens as __TOKEN_TYPE__
3) The size of execution tokens as __TOKEN_SIZE__
4) An optional token predecoder as __TOKEN_PREDECODER__
5) .....


Other parameters:
WORD_ALIGN_NEEDED (should be replaced by alignment constraint for each size)
"

"
------ Preferences ------

1) Intruction Type
   Possible values are CASE and FUNC. Previously included modules may have
defined a replacement (e.g. module gcc defines LABEL)
"

(define-default **INSTRUCTION-TYPE** 'CASE)

"
2) Token Type
   Choose whichever you prefer, and copy it in first position.

   DIRECT is fast, but generates large code.
   INDIRECT is not as fast, but more portable.

	defdef({__TOKEN_TYPE__},{DIRECT})

3) Token size
   WORD is the fastest, but takes some place
   BYTE is the smallest, but slower; needs a lot of (unimplemented) hack
        to work as direct.
   HALFWORD is intermediate, but does not work on all processors.

	defdef({__TOKEN_SIZE__},{WORD})

4) Token multiplier and offset

   No encoding is the default.
	defdef({__TOKEN_PREDECODER__},{$1})

   Another typical way to do things:
	defdef({__TOKEN_MULTIPLIER__},1)
	defdef({__TOKEN_OFFSET__},0)
	defdef({__TOKEN_PREDECODER__},
		{((($1)*(__TOKEN_MULTIPLIER__)+(__TOKEN_OFFSET__)))})


-----------------------------------------------------------------------------
----------------------- Defining instruction type ---------------------------
-----------------------------------------------------------------------------
   Let xxx be the name of your threading method, as acquired from
__INSTRUCTION_TYPE__, then defining the threading method consists in defining
the __INSTR_xxx__yyy__ macros, whose meaning is, according to yyy:

yyy:		meaning:
HEADERS		called in diversion -1.
		Must do all initializing and terminating stuff
		about the module. Use {__HEADER__}, {__FOOTER__}, and suches.
		Done before tokenizing header
		$1: unique name/prefix to use for global variables in this
		instance of the interpreter
		$2: particular suggested name postfix for the interpreter
INIT		instruction to begin using the threading method
		$1,$2: same as for HEADERS
EXIT		instruction body to stop using the threading method
		$1,$2: same as for HEADERS
type		the type of threading-level instruction tokens
		$1: same as for HEADER
decl		a macro to declare an individual instruction, called in CODE
		$1: same as for HEADER
		$2: name of the instruction
		$3: body of the instruction
val		a macro to compute the usable instruction from its name
		$1: same as for HEADER
		$2: name of the instruction, as given to *_decl
enter		asm instruction to issue in an instruction body to prepare
		execution on next instruction.
next		asm instruction to issue in an instruction body to continue
		execution on next instruction.
next_follows	defined to 1 iff the absence of next causes the next
		instruction declared to be executed, else defined to 0
does_forward	defined to 1 iff PRE_GET_NEXT_INSTR gets forwarded before
		the instruction gets executed

   These macros can send data in any appropriate non-default segment using the
usual system. Any other macro can be defined. They can use these that will
be defined in that context:

name:		meaning:
__NEXT_INSTR__	value of the next instruction to execute,
		as given by *_val and computed by the execution
		token interpreter.
__PRE_GET_NEXT_INSTR__
		macro to execute to compute the next instruction,
		before any PC-relative computation is done
__GET_NEXT_INSTR__
		macro to execute just before to use __NEXT_INSTR__




------ Let's create macros for them ------

define({__DEF_INSTR_MACRO__},
  {define({INSTR_$1},{__INSTR_}__INSTRUCTION_TYPE__{__$1__}(${@}))})

__DEF_INSTR_MACRO__(HEADERS)
__DEF_INSTR_MACRO__(INIT)
__DEF_INSTR_MACRO__(EXIT)
__DEF_INSTR_MACRO__(type)
__DEF_INSTR_MACRO__(decl)
__DEF_INSTR_MACRO__(val)
__DEF_INSTR_MACRO__(enter)
__DEF_INSTR_MACRO__(next)
__DEF_INSTR_MACRO__(next_follows)
__DEF_INSTR_MACRO__(does_forward)



----------------- interpreted code using a switch statement ------------------

define({$$INSTR_CASE_LIST})
define({__INSTR_CASE__INIT__},		{$1$2 ();})
define({__INSTR_CASE__EXIT__},		{goto $1__exit;})
define({__INSTR_CASE__HEADER__},
{
  __HEADER__({
  typedef enum $1__Instruction
  BEGIN indir($$INSTR_CASE_LIST) END $1__Instr;
  void $1$2 (void)
  BEGIN
  })

  __CODE__({
    for(;;)
    BEGIN
      __GET_NEXT_INSTR__
      switch(__NEXT_INSTR__)
      BEGIN
  })

  __ENDCODE__({
        default:
           Abort("Invalid instruction.") ;
      END
      __PRE_GET_NEXT_INSTR__
    END
  })

  __FOOTER__({
  END
  })
})
define({__INSTR_CASE__type__},	{$1__Instr})
define({__INSTR_CASE__decl__},
{AppendVar($$INSTR_CASE_LIST,{,$1$2})
      case $1$2:
         $3
})
define({__INSTR_CASE__val__},	{$1$2})
define({__INSTR_CASE__enter__},		{})
define({__INSTR_CASE__next__},		{break;})
define({__INSTR_CASE__next_follows__},	{0})
define({__INSTR_CASE__does_forward__},	{0})

------------------- interpreted code using a function call -------------------

defdef({__INSTR_FUNC__args__},		{})
defdef({__INSTR_FUNC__argsdecl__},	{void})
define({__INSTR_FUNC__INIT__},		{$1$2 ();})
define({__INSTR_FUNC__HEADERS__},{
  __INIT__({
  void $1__init(void)
  BEGIN
  })

  __INITFOOT__({
  END
  })

  __EXITHEAD__({
  void $1__exit(void)
  BEGIN
  })

  __EXIT__({
  END
  })

  __FOOTER__({
  void $1$2 (void)
  BEGIN
    __NLEXIT_HEADER__($1)
    $1__init ();
    while (1) BEGIN
      __GET_NEXT_INSTR__
      (*(__NEXT_INSTR__))(__INSTR_FUNC_args__);
      __PRE_GET_NEXT_INSTR__
      __NLEXIT_TEST__($1)
    END
    __NLEXIT_FOOTER__($1)
    $1__exit ();
  END
  })
})
define({__INSTR_FUNC__type__},	{void (*)(__INSTR_FUNC__argsdecl__)})
define({__INSTR_FUNC__decl__},
{
void $1$2 (__INSTR_FUNC__argsdecl__)
BEGIN
  $3
END
})
define({__INSTR_FUNC__val__},	{$1$2})
define({__INSTR_FUNC__enter__},		{})
define({__INSTR_FUNC__next__},		{return;})
define({__INSTR_FUNC__next_follows__},	{0})
define({__INSTR_FUNC__does_forward__},	{0})
define({__INSTR_FUNC__EXIT__},		{__NLEXIT_GO__($1)})




-----------------------------------------------------------------------------
---------------------------- Execution tokens -------------------------------
-----------------------------------------------------------------------------


ifdef({WORD_ALIGN_NEEDED},{
ifelse(
  ifelse(__TOKEN_TYPE__,HALFWORDCODE,1,__TOKEN_TYPE__,BYTECODE,1,0),1,
  {_OUT_(
{I won't use byte or wordcode on a computer with alignment constraints for now,
'cause it's a pain to program efficiently. Using DIRECT code instead.
})define({__TOKEN_TYPE__},{DIRECT})})})


** FORWARD INCREMENTATION **
   If your CPU doesn't have post-increment addressing mode for PC,
then you can increment PC at the ENTER macro that begins instructions
instead of doing it during the NEXT macro. Doing it in NEXT is better
if your CPU can post-increment, not as good (but works fine, just a bit
slower) if it doesn't. Using pre-increment might be possible in further
version of Tunes.
   Increment forwarding only works if the threading method does forward
and defines INSTR_does_forward to 1.

defdef({PC_INCREMENT},{FORWARD})
defdef({PC_INCREMENT},{AUTO_POST})	post-increment
defdef({PC_INCREMENT},{PRE})		! doesn't have the same semantics !!!


------ Macros we rely upon ------
* the INSTR_* defined by threading.m4
* macros to register an instruction:


------ Macros we define ------
* the __*NEXT_INSTR__ macros used by threading.m4
* EXEC_* macros, similar to the INSTR_*


dnl define({__DEF_INSTR_MACRO__},
dnl   {define({INSTR_$1},{__INSTR_}__INSTRUCTION_TYPE__{__$1__}(${@}))})



------ Forward incrementation ------
we define

__NEXT_INSTR__	value of the next instruction to execute,
		as given by instr_val and computed by the execution
		token interpreter.
__PRE_GET_NEXT_INSTR__
		macro to execute to compute the next instruction,
		before any PC-relative computation is done
__GET_NEXT_INSTR__
		macro to execute just before to use __NEXT_INSTR__


ifelse(PC_INCREMENT:INSTR_does_forward,1:1,
{define({__PRE_GET_NEXT_INSTR__},{PROGRAM_COUNTER++;})
 define({__GET_NEXT_INSTR__},{;})
},
{define({__PRE_GET_NEXT_INSTR__},{})
 define({__GET_NEXT_INSTR__},{PROGRAM_COUNTER++;})
})
define({__NEXT_INSTR__},{INSTR_OF_TOKEN(*PROGRAM_COUNTER)})


------ Reading tokens ------

DeclareVariable({PROGRAM_COUNTER},{*Token_t})



------ Tokens to instructions ------


define({__TOKEN_DIRECT__type__},{INSTR_type})
define({__TOKEN_DIRECT__INSTR__},{$*})
define({__TOKEN_DIRECT__type__},{INSTR_type})
define({__TOKEN_DIRECT__INSTR__},{$*})
{__TOKEN_TYPE__},{INDIRECT},
	{},
{__TOKEN_TYPE__},{HALFWORDCODE},
	{},
{__TOKEN_TYPE__},{BYTECODE},
	{},
	{_OUTln_(Error: unknown token type)})
